/*
DL2 Monitor: ./pa/pachap.c
Copyright (C) 2002, Rolf Jger, Ulm (Germany)

See the file README for information on usage and redistribution
of this source code file, and for a DISCLAIMER OF ALL WARRANTIES.
*/

/*----------------------------------------------------------------------------
 * Datalink Monitor (pachap): Protocol Analysis (RFC1994)
 * Challenge Handshake Authentication Protocol
 *
 * Implementation Notes
 * P_* indicate frame positions of certain fields;
 * C_* indicate known values of the code field;
 * default contains: code, request-id
 * 
 * Author: Rolf Jger,Ulm
 *----------------------------------------------------------------------------
 * YYYYMMDD  History
 * 20020427  redefinition: packet_length as length.
 * 20020405  default selection extended by request-id.
 * 20020330  created from scratch
 * $Id: pachap.c,v 1.5 2002/04/29 12:59:20 dv Exp $
 *----------------------------------------------------------------------------
*/
# include <stdio.h>
# include <string.h>
# include "lib/dl2pa.h"
# include "pa/chap.h"

// CHAP Fram Positions:
# define P_CODE 0x00
# define P_REQUESTID 0x01
# define P_LENGTH 0x02
# define P_DATA 0x04
// {<length> <data>}, with 1 octet per <length> 

// known codes: 1..4
# define C_Challenge  0x01
# define C_Response   0x02
# define C_Success    0x03
# define C_Failure    0x04

// default indications:
static const char* codestr[] = {
  "",
  "Challenge",
  "Response",
  "Success",
  "Failure"
};
# define MAXCODE C_Failure

//-exports--------------------------------------------------------------------
int dl2_pachap(Dl2_Frame *frame,int verbose) {
  Octet *buffer = frame->buffer;
  int length = frame->length, i = 0;
  Octet code;
  Doctet pl;  

  // check for required packet length:
  if (length < P_DATA) return -1;

  //code:
  code = buffer[i];
  if (verbose & CHAP_Mcode) {
    printf("%s=%02x",CHAP_Fcode,code);
  } 
  else if (verbose & CHAP_Mdefault) {
    const char *cs = (code <= MAXCODE ? codestr[code] : PA_FC_UNKNOWN); 
    printf(" %s",codestr[code]); 
  } i = P_REQUESTID;

  // request-id:
  if (verbose & CHAP_Mrequest_id) {
    printf(" %s=%02x",CHAP_Frequest_id,buffer[i]);
  }
  else if (verbose & CHAP_Mdefault) {
    printf("(%02x)",buffer[i]);
  } i = P_LENGTH;

  // packet-length:
  if (verbose & CHAP_Mlength) {
    printf(" %s=",CHAP_Flength);
    for (; i < P_DATA; i++) printf("%02x",buffer[i]);
  } else i = P_DATA;

  // data:
  if ((pl = str2doct(buffer+P_LENGTH)) > length) pl = length;
  if (verbose & CHAP_Mdata) {
    printf(" %s=",CHAP_Fdata);
    for (; i < pl; i++) printf("%02x",buffer[i]);
  } else i = pl;

  if (code==C_Challenge || code==C_Response) {

    // value:
    Doctet vl;
    i = P_DATA+1; vl = i + buffer[P_DATA];
    if (verbose & CHAP_Mvalue) {
      printf(" %s=",CHAP_Fvalue);
      for (; i < vl; i++) printf("%02x",buffer[i]);
    } else i = vl;

    // name:
    if (verbose & CHAP_Mname) {
      printf(" %s=",CHAP_Fname);
      for (; i < pl; i++) printf("%02x",buffer[i]);
    } else i = pl;
  }
  else if (code==C_Success || code==C_Failure) {

    // message:
    if (verbose & CHAP_Mmessage) {
      printf(" %s=",CHAP_Fmessage);
      for (i = P_DATA; i < pl; i++) printf("%02x",buffer[i]);
    } i = pl;
  }
  return i;
} //dl2_pachap


//-locals---------------------------------------------------------------------

/***EOM***/
