/*
DL2 Monitor: ./pa/paicmp.c
Copyright (C) 2002, Rolf Jger, Ulm (Germany)

See the file README for information on usage and redistribution
of this source code file, and for a DISCLAIMER OF ALL WARRANTIES.
*/

/*----------------------------------------------------------------------------
 * Datalink Monitor (paicmp): Protocol Analysis (RFC792)
 *
 * Implementation Notes:
 * default contains: type, code;
 * data precedes any other interpretations;
 * otherwise, contents wil be interpreted depending on type, only;
 * type-dependent parameter interpretation has been bound to functions; 
 * 
 * Author: Rolf Jger,Ulm
 *----------------------------------------------------------------------------
 * YYYYMMDD  History
 * 20020501  fully implemented according to rfc792.
 * 20020330  created from scratch
 * $Id: paicmp.c,v 1.4 2002/05/01 18:57:35 dv Exp $
 *----------------------------------------------------------------------------
*/
# include <stdio.h>
# include <string.h>
# include "lib/dl2pa.h"
# include "pa/icmp.h"


// ICMPv4 frame positions:
# define P_TYPE      0x00
# define P_CODE      0x01
# define P_CHECKSUM  0x02
# define P_DATA      0x04

// known types: 0..16 (RFC792) 
# define T_EchoReply    0x00
# define T_DestUnreach  0x03
# define T_SourceQuench 0x04
# define T_Redirection  0x05
# define T_EchoRequest  0x08
# define T_TimeExceeded 0x0b
# define T_ParameterPrb 0x0c
# define T_TimeRequest  0x0d
# define T_TimeReply    0x0e
# define T_InfoRequest  0x0f
# define T_InfoReply    0x10
# define N_TYPES        (T_InfoReply+1)
# define UNKTYPE        N_TYPES

// known parameters/arguments depending on type:
# define A_orgdgrm      ICMP_Morg_datagram
# define A_pointer      ICMP_Mpointer
# define A_gateway      ICMP_Mgateway_addr
# define A_identifier   ICMP_Mrequest_id
# define A_sequence     ICMP_Msequence_no
# define A_orgtime      ICMP_Morg_timestamp
# define A_rcvtime      ICMP_Mrcv_timestamp
# define A_trmtime      ICMP_Mtrm_timestamp
# define A_echodata     ICMP_Mecho_data

// default indications, indexed by type:
typedef int (*Printfct)(int,Octet*,int);
typedef struct Icmpinfo {
  const char *name;
  Printfct print_message;
  int parameter;
} Icmpinfo;  

static int print_orgdgrm(int,Octet*,int);
static int print_echo(int,Octet*,int);
static int print_time(int,Octet*,int);
static int print_info(int,Octet*,int);
static int print_redirect(int,Octet*,int);
static int print_parameter(int,Octet*,int);

# define ADD_ICMPINFO(t,f,m) [T_##t] {#t, print_##f, m}
static Icmpinfo icmpinfo[N_TYPES+1] = {
  ADD_ICMPINFO(EchoReply,   echo,     A_identifier|A_sequence|A_echodata),
  ADD_ICMPINFO(EchoRequest, echo,     A_identifier|A_sequence|A_echodata),
  ADD_ICMPINFO(DestUnreach, orgdgrm,  A_orgdgrm),
  ADD_ICMPINFO(SourceQuench,orgdgrm,  A_orgdgrm),
  ADD_ICMPINFO(Redirection, redirect, A_gateway|A_orgdgrm),
  ADD_ICMPINFO(TimeExceeded,orgdgrm,  A_orgdgrm),
  ADD_ICMPINFO(ParameterPrb,parameter,A_pointer|A_orgdgrm),
  ADD_ICMPINFO(TimeReply,   time,     A_identifier|A_sequence|A_orgtime|A_rcvtime|A_trmtime),
  ADD_ICMPINFO(TimeRequest, time,     A_identifier|A_sequence|A_orgtime|A_rcvtime|A_trmtime),
  ADD_ICMPINFO(InfoReply,   info,     A_identifier|A_sequence),
  ADD_ICMPINFO(InfoRequest, info,     A_identifier|A_sequence),
  [UNKTYPE] {PA_FT_UNKNOWN,NULL,0}
};

 
//-exports--------------------------------------------------------------------
int dl2_paicmp(Dl2_Frame *frame,int verbose) {
  Octet *buffer = frame->buffer;
  int length = frame->length, i = P_TYPE;
  Octet type;
  Icmpinfo *pi = &icmpinfo[UNKTYPE]; 

  // check for required packet length:
  if (length < P_DATA) return -1;

  // type:
  type = buffer[i];
  if (type < N_TYPES) pi = &icmpinfo[type];
  if (verbose & ICMP_Mtype) {
    printf("%s=%02x",ICMP_Ftype,buffer[i]);
  } 
  else if (verbose & ICMP_Mdefault) {
    printf(" %s",pi->name); 
  } i = P_CODE;

  // code:
  if (verbose & (ICMP_Mcode | ICMP_Mdefault)) {
    printf(" %s=%02x",ICMP_Fcode,buffer[i]);
  } i = P_CHECKSUM;

  // checksum:
  if (verbose & ICMP_Mchecksum) {
    printf(" %s=%02x%02x",ICMP_Fchecksum,buffer[i],buffer[i+1]);
  } i = P_DATA;
 
  // data:
  if (verbose & ICMP_Mdata) {
    printf(" %s=",ICMP_Fdata);    
    for (;i < length; i++) printf("%02x",buffer[i]);
  }

  // detailed interpretation:
  else if ((type < N_TYPES) && (pi->print_message != 0)) {
    int vbflags = (pi->parameter & verbose);
    i += pi->print_message(vbflags,buffer+i,length-i);
  } 
  else i = length;

  return i;
} //dl2_paicmp


//-locals---------------------------------------------------------------------
/* print an ICMP message depending on type, return number of analyzed octets;
 * buffer points to first data octet, ml indicates the message length;
 * message format: see rfc792;
*/
inline int print_orgdgrm(int vb,Octet *buffer,int ml) {
  int i;
  if (vb & ICMP_Morg_datagram) {
    printf(" %s=",ICMP_Forg_datagram);    
    for (i = 4; i < ml; i++) printf("%02x",buffer[i]); 
  } return ml;
} //print_orgdgrm


inline int print_echo(int vb,Octet *buffer,int ml) {
  int i;
  if (ml >= 4) {
    if (vb & ICMP_Mrequest_id) {
      printf(" %s=%02x%02x",ICMP_Frequest_id,buffer[0],buffer[1]);    
    } 
    if (vb & ICMP_Msequence_no) {
      printf(" %s=%02x%02x",ICMP_Fsequence_no,buffer[2],buffer[3]);    
    } 
    if (vb & ICMP_Mecho_data) {
      printf(" %s=",ICMP_Fecho_data);
      for (i = 4; i < ml; i++) printf("%02x",buffer[i]);     
    } 
  } return ml;
} //print_echo


inline int print_time(int vb,Octet *buffer,int ml) {
  int i;
  if (ml >= 4) {
    if (vb & ICMP_Mrequest_id) {
      printf(" %s=%02x%02x",ICMP_Frequest_id,buffer[0],buffer[1]);    
    } 
    if (vb & ICMP_Msequence_no) {
      printf(" %s=%02x%02x",ICMP_Fsequence_no,buffer[2],buffer[3]);    
    } i = 4;
    if (vb & ICMP_Morg_timestamp) {
      printf(" %s=",ICMP_Forg_timestamp);
      for (; i < 8; i++) printf("%02x",buffer[i]);     
    } i = 8;
    if (vb & ICMP_Mrcv_timestamp) {
      printf(" %s=",ICMP_Frcv_timestamp);
      for (; i < 12; i++) printf("%02x",buffer[i]);     
    } i = 12;
    if (vb & ICMP_Mtrm_timestamp) {
      printf(" %s=",ICMP_Ftrm_timestamp);
      for (; i < 16; i++) printf("%02x",buffer[i]);     
    } i = 16;
    return i;
  } return ml;
} //print_time


inline int print_info(int vb,Octet *buffer,int ml) {
  int i;
  if (ml >= 4) {
    ml = 4;
    if (vb & ICMP_Mrequest_id) {
      printf(" %s=%02x%02x",ICMP_Frequest_id,buffer[0],buffer[1]);    
    } 
    if (vb & ICMP_Msequence_no) {
      printf(" %s=%02x%02x",ICMP_Fsequence_no,buffer[2],buffer[3]);    
    } 
  } return ml;
} //print_info


inline int print_redirect(int vb,Octet *buffer,int ml) {
  int i;
  if (ml >= 4) {
    if (vb & ICMP_Mgateway_addr) {
      printf(" %s=%02x%02x%02x%02x",ICMP_Fgateway_addr,
        buffer[0],buffer[1],buffer[2],buffer[3]);
    }
    if (vb & ICMP_Morg_datagram) {
      printf(" %s=",ICMP_Forg_datagram);    
      for (i = 4; i < ml; i++) printf("%02x",buffer[i]); 
    }
  } return ml;
} //print_redirect


inline int print_parameter(int vb,Octet *buffer,int ml) {
  int i;
  if (ml >= 4) {
    if (vb & ICMP_Mpointer) {
      printf(" %s=%02x",ICMP_Fpointer,buffer[0]);
    }
    if (vb & ICMP_Morg_datagram) {
      printf(" %s=",ICMP_Forg_datagram);    
      for (i = 4; i < ml; i++) printf("%02x",buffer[i]); 
    }
  } return ml;
} //print_parameter

/***EOM***/
